/*
*/

import java.io.*;
import java.net.*;
import java.lang.*;
import java.applet.*;
import vrml.external.field.*;
import vrml.external.exception.*;
import vrml.external.Node;
import vrml.external.Browser;
import Chess;

public abstract class 
Piece extends Object implements EventOutObserver 
{
  public static final int REAL  = 0;
  public static final int PROXY = 1;

  public static final int KILLED  = 0;
  public static final int PICKED  = 1;
  public static final int FREE    = 2;

  protected String   dataFile = null;
  protected String   scaleFile = null;
  protected float    scaleInfo = 1f;
  protected Position initialPosition;
  protected Position currentPosition;
  protected char     color;
  protected short    id;
  protected String   name = null;

  // Is it real or proxy
  protected int      type = 0;

  // Is it picked/killed/free 
  protected int      state = 0;

  // CosmoPlayer we're using to instantiate these nodes
  protected Browser browser;
  

  private static String King  = "null";
  private static String Queen = "null";
  private static String Bishop = "null";
  private static String Knight = "null";
  private static String Rook = "null";
  private static String Pawn = "null";
  public static float kingScale;
  public static float queenScale;
  public static float bishopScale;
  public static float knightScale;
  public static float rookScale;
  public static float pawnScale;
  public static String filehead = "null";
  private final static String Square = "DEF tile IndexedFaceSet {ccw FALSE " +
"			coord Coordinate { point [12 0 12,-12 0 12,-12 0 -12,12 0 -12]} "+
"			coordIndex [2, 3, 0, -1, 1, 2, 0, -1] " +
"		} ";
	  

  
  Chess parent = null;

  private static boolean first = true;
  
  // Nodes we're interested in
  public Node[] transArray;
  Node transform;
  Node material;
  Node PositionInterpolator;
  
  // EventIns we'll modify
  public EventInSFRotation set_rotation;
  public EventInSFVec3f    set_scale;
  public EventInSFVec3f    set_translation;
  public EventInSFColor    set_diffuseColor;
  public EventInSFColor    set_specularColor;
  public EventInSFVec2f    set_shininess;
  public EventInMFVec3f    set_keyValue;
  public EventInSFTime     set_startTime;
  public Node pos;
  public Node time;
  
  // EventOuts we'll query
  public EventOutSFRotation rotation_changed;
  public EventOutSFVec3f    scale_changed;
  public EventOutSFVec3f    translation_changed;
  public EventOutSFColor    diffuseColor_changed;
  public EventOutSFTime     touchTime_changed;
  
  // State associated with the board matrix
  public int xang;
  public int yang;

  // contact between board and piece
  public boolean killed = false;
  public boolean moved = false;
  public double lastTouch = 0;
  public Piece where = null;
  public float x;
  public float y;
  public float z;
  public int num = -1;
  
  private int pieceType = -1;

  private static Piece previous = null;

  public Piece() {}
  
  Piece(Chess myParent, int whichType, Piece Where, float zz, int Num) {
    num = Num;
    where = Where;
    Where.where = this;
    this.x = Where.x;
    this.y = Where.y;
    this.xang = Where.xang;
    this.yang = Where.yang;
    this.z = zz;
    this.preparePiece(myParent, whichType);
  }
  

  Piece(Chess myParent, int whichType, float xx, float yy, int xxx, int yyy) {
    x = xx;
    y = yy;
    xang = xxx;
    yang = yyy;
    this.preparePiece(myParent, whichType);
  }

  public void smoothTranslation(float xx, float yy, float zz, double t) {
    float[][] f = new float[2][3];
    f[0][0] = this.x; f[0][1] = this.z; f[0][2] = this.y; // source
    f[1][0] = xx;     f[1][1] = yy;     f[1][2] = zz; // destination
    set_keyValue.setValue(f);
    previous = this;
    set_startTime.setValue(t);
    this.x = xx; this.z = yy; this.y = zz;
  } 

  private void 
  definePiece(Chess myParent, int whichType) 
  {
    parent = myParent;
    browser = parent.getBrowser();
    if (filehead.equals("null")) {
      URL temp_URL = parent.getCodeBase();
      filehead = temp_URL.toString();
    }

    if ((whichType != Chess.W_SQUARE)&&(whichType != Chess.B_SQUARE)) 
    {
      pos = browser.getNode("Pos" + String.valueOf(num));
      time = browser.getNode("Time" + String.valueOf(num));
    }
    
    try 
    {
      // Instantiate common nodes

      /*
      if ((whichType != Chess.W_SQUARE)&&(whichType != Chess.B_SQUARE)) 
	    transform = browser.getNode("A" + String.valueOf(num));
      transArray = browser.createVrmlFromString("DEF A" + String.valueOf(num) + " Transform {}");
      if (num == -1) transArray = browser.createVrmlFromString("Transform {}");
      */ 

	  transform = browser.getNode("GAME");
      transArray = browser.createVrmlFromString("DEF "+ getPieceHashKey() + " Transform {}");

      Node[] shapeArray = browser.createVrmlFromString("Shape {}");
      Node[] matArray = browser.createVrmlFromString("Material {}");
      Node[] appArray = browser.createVrmlFromString("Appearance {}");
      Node[] sensArray = browser.createVrmlFromString("TouchSensor {}");
      Node[] sensArray2 = browser.createVrmlFromString("PlaneSensor {}");
      Node[] geomArray = null;
      pieceType = whichType;
      Float t;
      String ss;
      
      // Read the Definition Files
	  String definition = this.readText(filehead + dataFile);
	  String sclInfo = this.readText(filehead + scaleFile);
	  ss = sclInfo.substring(0, sclInfo.indexOf('f'));
	  t = new Float(ss);
	  pawnScale = t.floatValue();
	  scaleInfo = t.floatValue();
	  geomArray = browser.createVrmlFromString(definition);

      //if ((whichType == Chess.B_SQUARE)||(whichType == Chess.W_SQUARE)) {
	  //geomArray = browser.createVrmlFromString(Square); }
      
      // Put together the hierarchy
      //if ((whichType == Chess.W_SQUARE)||(whichType == Chess.B_SQUARE)) 
      //    transform = transArray[0];

      material = matArray[0];
      EventInSFNode nodeIn = 
                    (EventInSFNode) shapeArray[0].getEventIn("appearance"); 
      nodeIn.setValue(appArray[0]); 

      nodeIn = (EventInSFNode) appArray[0].getEventIn("material");
      nodeIn.setValue(material); 

      nodeIn = (EventInSFNode) shapeArray[0].getEventIn("geometry"); 
      nodeIn.setValue(geomArray[0]); 

      EventInMFNode nodesIn =
                   (EventInMFNode) transform.getEventIn("addChildren"); 

      nodesIn.setValue(shapeArray); 
      nodesIn.setValue(sensArray); 
      nodesIn.setValue(sensArray2); 
      
      // Extract EventIns and EventOuts of transform and material node
      set_rotation = (EventInSFRotation) transform.getEventIn("rotation");
      set_scale = (EventInSFVec3f) transform.getEventIn("scale");
      set_translation = (EventInSFVec3f) transform.getEventIn("translation");
      set_diffuseColor = (EventInSFColor) material.getEventIn("diffuseColor");
      set_specularColor = (EventInSFColor) material.getEventIn("specularColor");

      /*
      if ((whichType != Chess.W_SQUARE)&&(whichType != Chess.B_SQUARE)) 
      {
	     set_keyValue = (EventInMFVec3f) pos.getEventIn("keyValue");
	     set_startTime = (EventInSFTime) time.getEventIn("startTime");
      }
      */
	     set_keyValue = (EventInMFVec3f) pos.getEventIn("keyValue");
	     set_startTime = (EventInSFTime) time.getEventIn("startTime");
      
      rotation_changed = (EventOutSFRotation) transform.getEventOut("rotation");
      scale_changed = (EventOutSFVec3f) transform.getEventOut("scale");
      translation_changed =
             (EventOutSFVec3f) transform.getEventOut("translation");

      diffuseColor_changed = 
	         (EventOutSFColor) material.getEventOut("diffuseColor");

      touchTime_changed =
	         (EventOutSFTime) sensArray[0].getEventOut("touchTime");
      
      // Set EventOut callback for this clump's touch sensor,
      // so we can later read in its values when it's clicked
      touchTime_changed.advise(this, null);

    }
    
    catch (InvalidVrmlException e) {
      System.out.println("PROBLEMS!: " + e);
    }
    
    catch (InvalidEventInException e) {
      System.out.println("PROBLEMS!: " + e);
    }
    
    catch (InvalidEventOutException e) {
      System.out.println("PROBLEMS!: " + e);
    }
    first = false;

  }
  
  

  private void 
  preparePiece(Chess myParent, int whichType) 
  {
    parent = myParent;
    browser = parent.getBrowser();
    if (filehead.equals("null")) {
      URL temp_URL = parent.getCodeBase();
      filehead = temp_URL.toString();
    }

    if ((whichType != Chess.W_SQUARE)&&(whichType != Chess.B_SQUARE)) 
    {
      pos = browser.getNode("Pos" + String.valueOf(num));
      time = browser.getNode("Time" + String.valueOf(num));
    }
    
    try 
    {
      // Instantiate common nodes

      if ((whichType != Chess.W_SQUARE)&&(whichType != Chess.B_SQUARE)) 
	transform = browser.getNode("A" + String.valueOf(num));
      transArray = browser.createVrmlFromString("DEF A" + String.valueOf(num) + " Transform {}");
      if (num == -1)
	  transArray = browser.createVrmlFromString("Transform {}");

      Node[] shapeArray = browser.createVrmlFromString("Shape {}");
      Node[] matArray = browser.createVrmlFromString("Material {}");
      Node[] appArray = browser.createVrmlFromString("Appearance {}");
      Node[] sensArray = browser.createVrmlFromString("TouchSensor {}");
      Node[] geomArray = null;
      pieceType = whichType;
      Float t;
      String ss;
      
      if ((whichType == Chess.W_KING)||(whichType == Chess.B_KING)) {
	if (King.equals("null")) {
	  King = this.readText(filehead + "King.dat");
	  ss = this.readText(filehead + "kingScale.dat");
          ss = ss.substring(0, ss.indexOf('f'));
	  t = new Float(ss);
	  kingScale = t.floatValue();
	}
	geomArray = browser.createVrmlFromString(King);
      }
      if ((whichType == Chess.W_QUEEN)||(whichType == Chess.B_QUEEN)) {
	if (Queen.equals("null")) {
	  Queen = this.readText(filehead + "Queen.dat");
	  ss = this.readText(filehead + "queenScale.dat");
	  ss = ss.substring(0, ss.indexOf('f'));
	  t = new Float(ss);
	  queenScale = t.floatValue();
	}
	geomArray = browser.createVrmlFromString(Queen);
      }
      if ((whichType == Chess.W_BISHOP)||(whichType == Chess.B_BISHOP)) {
	if (Bishop.equals("null")) {
	  Bishop = this.readText(filehead + "Bishop.dat");
	  ss = this.readText(filehead + "bishopScale.dat");
	  ss = ss.substring(0, ss.indexOf('f'));
	  t = new Float(ss);
	  bishopScale = t.floatValue();
	}
	geomArray = browser.createVrmlFromString(Bishop);
      }
      if ((whichType == Chess.W_KNIGHT)||(whichType == Chess.B_KNIGHT)) {
	if (Knight.equals("null")) {
	  Knight = this.readText(filehead + "Knight.dat");
	  ss = this.readText(filehead + "knightScale.dat");
	  ss = ss.substring(0, ss.indexOf('f'));
	  t = new Float(ss);
	  knightScale = t.floatValue();
	}
	geomArray = browser.createVrmlFromString(Knight);
      }
      if ((whichType == Chess.W_ROOK)||(whichType == Chess.B_ROOK)) {
	if (Rook.equals("null")) {
	  Rook = this.readText(filehead + "Rook.dat");
	  ss = this.readText(filehead + "rookScale.dat");
	  ss = ss.substring(0, ss.indexOf('f'));
	  t = new Float(ss);
	  rookScale = t.floatValue();}
	geomArray = browser.createVrmlFromString(Rook);
      }
      if ((whichType == Chess.W_PAWN)||(whichType == Chess.B_PAWN)) {
	if (Pawn.equals("null")) {
	  Pawn = this.readText(filehead + "Pawn.dat");
	  ss = this.readText(filehead + "pawnScale.dat");
	  ss = ss.substring(0, ss.indexOf('f'));
	  t = new Float(ss);
	  pawnScale = t.floatValue();}
	geomArray = browser.createVrmlFromString(Pawn);
      }
      if ((whichType == Chess.B_SQUARE)||(whichType == Chess.W_SQUARE)) {
	geomArray = browser.createVrmlFromString(Square);
      }
      
      // Put together the hierarchy
      if ((whichType == Chess.W_SQUARE)||(whichType == Chess.B_SQUARE)) 
	transform = transArray[0];
      material = matArray[0];
      EventInSFNode nodeIn = 
                    (EventInSFNode) shapeArray[0].getEventIn("appearance"); 
      nodeIn.setValue(appArray[0]); 

      nodeIn = (EventInSFNode) appArray[0].getEventIn("material");
      nodeIn.setValue(material); 

      nodeIn = (EventInSFNode) shapeArray[0].getEventIn("geometry"); 
      nodeIn.setValue(geomArray[0]); 

      EventInMFNode nodesIn =
                   (EventInMFNode) transform.getEventIn("addChildren"); 

      nodesIn.setValue(shapeArray); 
      nodesIn.setValue(sensArray); 
      
      // Extract EventIns and EventOuts of transform and material node
      set_rotation = (EventInSFRotation) transform.getEventIn("rotation");
      set_scale = (EventInSFVec3f) transform.getEventIn("scale");
      set_translation = (EventInSFVec3f) transform.getEventIn("translation");
      set_diffuseColor = (EventInSFColor) material.getEventIn("diffuseColor");
      set_specularColor = (EventInSFColor) material.getEventIn("specularColor");
      if ((whichType != Chess.W_SQUARE)&&(whichType != Chess.B_SQUARE)) {
	set_keyValue = (EventInMFVec3f) pos.getEventIn("keyValue");
	set_startTime = (EventInSFTime) time.getEventIn("startTime");
      }
      
      rotation_changed = (EventOutSFRotation) transform.getEventOut("rotation");
      scale_changed = (EventOutSFVec3f) transform.getEventOut("scale");
      translation_changed =
	(EventOutSFVec3f) transform.getEventOut("translation");
      diffuseColor_changed = 
	(EventOutSFColor) material.getEventOut("diffuseColor");
      touchTime_changed =
	(EventOutSFTime) sensArray[0].getEventOut("touchTime");
      
      // Set EventOut callback for this clump's touch sensor,
      // so we can later read in its values when it's clicked
      touchTime_changed.advise(this, null);

    }
    
    catch (InvalidVrmlException e) {
      System.out.println("PROBLEMS!: " + e);
    }
    
    catch (InvalidEventInException e) {
      System.out.println("PROBLEMS!: " + e);
    }
    
    catch (InvalidEventOutException e) {
      System.out.println("PROBLEMS!: " + e);
    }
    first = false;

  }
  

  // We'll take care of handling the callback every time our
  // TouchSensor is clicked.
  public void callback(EventOut event, double time, Object userData) {
    // Make ourselves the current clump
    lastTouch = time;
    parent.makeCurrent(this);
  }
  

  public int Type() {
    return pieceType;
  }

  public String
  getPieceHashKey() 
  { 
     String ret = new String(getPieceColor()+getPieceName()+getPieceNumber());
     return ret; 

  }// end of getPieceHashKey()

  public String
  getPieceName() { return name; }

  public String
  getPieceColor() { return (color=='W'?"White":"Black"); }

  public int
  getPieceType() { return type; }

  public int
  getPieceNumber() { return id; }

  public Position
  getPieceCurrentPosition() { return currentPosition; }

  /**
  * This message has to be broadcasted to everybody.
  */
  public void
  setPieceCurrentPosition(Position newPosition) 
  { 
      currentPosition = newPosition;

      if(type==REAL)
      {
      }// end of if

  }// end of setPieceCurrentPosition()

  public int
  getPieceState() { return state; }

  /**
  * This message has to be broadcasted to everybody.
  */
  public void
  setPieceState(int newState) 
  { 
    if(newState>-1 && newState<3) 
    {
       if(state==FREE && newState==KILLED) state = KILLED ; 
       else if(state==PICKED && newState==FREE) state = FREE ; 
       else if(state==FREE && newState==PICKED) state = PICKED ; 

    }// end of if 

  }// end of setPieceState()

  public boolean
  canBePicked()
  {  
     boolean ret = (state==FREE);

     return ret;

  }// end of canBePicked()
  
  public abstract boolean canBePlacedAt(Position newPosition) ;

  private String 
  readText(String file)
  {
     String data = new String();
    
     try
     {
	    // URL for configuration file
	    URL url = new URL(file);
	
	    // if the file does not exist, the following will throw an exception: 
	
	    URLConnection urlConnection = url.openConnection();
	    urlConnection.connect();
	
	    int file_length = urlConnection.getContentLength();
	
	    ByteArrayOutputStream buffer = new ByteArrayOutputStream();
	    byte[] buf = new byte[file_length+1];
	    int read = 0;
	    int total_read = 0;
	
	    while(total_read!=file_length)
	    {
	      InputStream is = urlConnection.getInputStream();
	      read = is.read(buf);
	      buffer.write(buf,0,read);
	      total_read += read;  

	    }// end of while
	
	    data = buffer.toString();

      } catch (Exception ex) 
        { System.out.println("File doesn't exist"); }

    return data;

  }// end of readText()

}// end of Piece()
 

